unit Matr;

interface
uses jpeg; // for TJPEGQualityRange
var
    CompressionQuality: TJPEGQualityRange;
type
    matrix = class
    protected
      mx, my: integer;
      w: array of array of extended;
    public
      constructor Create;                       overload;
      constructor Create( mx_, my_: integer);   overload;
      Destructor  Done;

      function    getMx():integer;
      function    getMy():integer;
      procedure   testSameSize( u: matrix );
      procedure   setSize( mx_, my_: integer );
      procedure   copyFrom( p: matrix );

      function    isin(x,y: integer): boolean;
      function    get( x,y: integer): extended;   overload;
      //function    get( p: Ppoint): extended;      overload;
      function    getI( x,y: integer): integer;
      procedure   put( x,y: integer; v:extended);
      procedure   incr( x,y: integer; v:extended);	//    

      function    get1 (x,y: extended ):extended;    // bilinear interpolation
      function    get3 (x,y: extended ):extended;    // bicubic interpolation

      procedure   SaveToTextFile   (fname: string); overload; // Save to text file
      procedure   SaveToTextFile   (fname: string; l:integer); overload;  // Save to text file
      procedure   SaveToTextFileXY (fname: string; l:integer);// Save to text file
      procedure   SaveIToTextFile  (fname: string); // Save integers to text file
      procedure   SaveIToTextFileXY(fname: string); // Save integers to text file

      procedure   LoadFromTextFile(fname: string);   // Load from text file

      procedure   LoadGreyFromFile(fname: string);   // w[y,x] := ITU-601(red,green,blue)
      procedure   LoadGreyFromFile_(fname:string);   // w[y,x] := (red + green + blue)/3
      procedure   SaveGreyToFile  (fname: string);

      // algebra
      procedure   setValue( u: extended );            // w[*.*] := u;
      procedure   addValue( u: extended );            // w[*.*] += u;
      procedure   mulValue( u: extended );            // w[*.*] *= u;

      procedure   add( v: matrix );
      procedure   sub( v: matrix );
      procedure   addC( v: matrix; k: extended );     // self += v*u
      procedure   mul( v: matrix );                   //  
      //
      function    min():extended;                 overload;
      function    min(var x,y: integer):extended; overload;
      function    max():extended;                 overload;
      function    max(var x,y: integer):extended; overload;

      procedure   clip( k0, k1: extended );    //      [k0,k1]
      procedure   norm( k0, k1: extended );    //      [k0,k1]
      procedure   roundI();                    //    

      procedure   smoothSlow(n: integer);      //  :    -     (2k+1)*(2k+1)
      procedure   smooth(n: integer);          //    -     (2k+1)*(2k+1)
      procedure   smooth2(n: integer);         //    -    ....

      function    dist2( a: matrix ):extended; //  

      // matrix as mask ------------------------------------------------------------
      //    [x,y],   a.w[y,x]=<v,  self.w[y,x]:=m
      function    setLess( a: matrix; v, m: extended ):integer;

      //    [x,y],   a.w[y,x]>=v,  self.w[y,x]:=m
      function    setGreater( a: matrix; v, m: extended ):integer;

      //    [x,y],   v1<=a.w[y,x]<=v2,  self.w[y,x]:=m
      function    setInterval( a: matrix; v1, v2, m: extended ):integer;

      function    countValue( m: extended ):integer; //     
      function    changeValue( m1, m2: extended ):integer; //  m1     m2.   

	  // 
      procedure   diffX ( a: matrix );			// this := da/dx   
      procedure   diffY ( a: matrix );			// this := da/dy   
      procedure   diffXX( a: matrix );          // this := da/dxdx   
      procedure   diffXY( a: matrix );          // this := da/dxdy   
      procedure   diffYY( a: matrix );          // this := da/dydy   

	  // resize "A"
	  procedure	  downSampl_A (a: matrix; times: integer; align: integer=1); // this:=downsampe(a) by times
	  procedure	  resizeBilin_A(a: matrix; mx_, my_: integer); // this:=resize(a) to (mx_, my_)
	  procedure	  resizeBicub_A(a: matrix; mx_, my_: integer); // this:=resize(a) to (mx_, my_)

	  // resize "B"
	  procedure	  downSampl (a: matrix; times: integer; align: integer=1); // this:=downsampe(a) by times
	  procedure	  downSamplI(a: matrix; times: integer; align: integer=1); // this:=downsampe(a) by times
	  procedure	  resizeBilin(a: matrix; mx_, my_: integer); // this:=resize(a) to (mx_, my_)
	  procedure	  resizeBicub(a: matrix; mx_, my_: integer); // this:=resize(a) to (mx_, my_)

      	// "B":         
	  procedure	  expand2_Base (a: matrix); //  
	  procedure	  expand2_Bilin(a: matrix); //  
	  procedure	  expand2_Bicub(a: matrix); //  
	  procedure	  expand2_4Point(m: matrix; al,be,ga:extended); // 4-    (al,be,ga,1-al-be-ga)

    end;

procedure LoadFromFile(r,g,b: matrix; fname: string);
procedure SaveToFile  (r,g,b: matrix; fname: string);

implementation
uses graphics, SysUtils, classes, Math,			// standart
     common;									// additional
// ========================================================================

// ========================================================================
// class matrix
// ========================================================================
constructor matrix.Create;
begin
  mx := 0; my := 0; w:=nil;
end;
//-------------------------------------------------------------------------
constructor matrix.Create( mx_, my_: integer);
begin
  if (mx_<1) or (mx_>1000000) or (my_<1) or (my_>1000000) then
  	ErrProc('matrix.Create('+I2Str(mx_,8)+','+I2Str(my_,8)+')');
  my := my_; mx := mx_;
  SetLength(w, my, mx);
end;
//-------------------------------------------------------------------------
Destructor  matrix.Done; begin
	mx:=0;
    my:=0;
    w:=nil;
end;
//-------------------------------------------------------------------------
function    matrix.getMx():integer; begin getMx:=mx; end;
function    matrix.getMy():integer; begin getMy:=my; end;
//-------------------------------------------------------------------------
procedure   matrix.testSameSize( u: matrix );
begin
    if u.mx <> mx then ErrProc('testSameSize');
    if u.my <> my then ErrProc('testSameSize');
end;
// -------------------------------------------------------------
procedure   matrix.setSize( mx_, my_: integer );
begin
	if (mx_<1) or (mx_>1000000) or (my_<1) or (my_>1000000) then
  		ErrProc('matrix.setSize('+I2Str(mx_,8)+','+I2Str(my_,8)+')');
    if (mx=mx_) and (my=my_) then exit;
    mx:=mx_; my:=my_;
    SetLength(w, my, mx);
end;
// -------------------------------------------------------------
procedure   matrix.copyFrom( p: matrix );
var x,y: integer;
begin
    self.setSize(p.getMx,p.getMy);
    for y:=0 to my-1 do
    for x:=0 to mx-1 do w[y,x]:=p.w[y,x];
end;
// -------------------------------------------------------------
function    matrix.isin( x,y: integer): boolean;
begin
  isin := (y>=0) and (y<my) and (x>=0) and (x<mx);
end;
// -------------------------------------------------------------
function    matrix.get( x,y: integer): extended;
begin
  if y<  0 then y:=0;
  if y>=my then y:=my-1;
  if x<  0 then x:=0;
  if x>=mx then x:=mx-1;
  get := w[y,x];
end;
// -------------------------------------------------------------
{
function    matrix.get( p: Ppoint): extended;
var x,y: integer;
begin
  x:=p.x;
  y:=p.y;
  if y<  0 then y:=0;
  if y>=my then y:=my-1;
  if x<  0 then x:=0;
  if x>=mx then x:=mx-1;
  get := w[y,x];
end;
}
// -------------------------------------------------------------
function    matrix.getI( x,y: integer): integer;
begin
  getI := round( get(x,y) );
end;

// -------------------------------------------------------------
procedure    matrix.put( x, y: integer; v:extended);
begin
  if (y>=0) and (y<my) and (x>=0) and (x<mx) then w[y,x]:=v;
end;
// -------------------------------------------------------------
procedure    matrix.incr( x, y: integer; v:extended);	//    
begin
  if (y>=0) and (y<my) and (x>=0) and (x<mx) then w[y,x]:=w[y,x]+v;
end;
// -------------------------------------------------------------
procedure matrix.LoadGreyFromFile(fname: string);// w[y,x] := ITU-601(red,green,blue)
var BitMap : TBitMap;
    jp     : TJpegImage;
    x,y    : integer;
    ext    : string[5];
    pl     : PByteArray;
begin
  BitMap := TBitMap.Create;
  jp := TJpegImage.Create;

  ext := UpperCase( ExtractFileExt(fname));
  if (ext <> '.BMP') and (ext <> '.JPG') and (ext <> '.JPEG') then Exit;

  try
    if ext= '.BMP' then BitMap.LoadFromFile(fname)
    else begin
      jp.LoadFromFile(fname);
      BitMap.Assign(jp);
    end;
    my:=Bitmap.Height;
    mx:=Bitmap.Width;
    SetLength(w, my, mx);
    {     : Bitmap -> (pr, pg, pb)}
    for y:= 0 to my-1 do begin
      pl := BitMap.ScanLine[y];
      for x:= 0 to mx-1 do
      //w[y,x] := integer(pl[3*x  ]) + pl[3*x+1] + pl[3*x+2];
      w[y,x] := (19595*integer(pl[3*x+2]) +38470*integer(pl[3*x+1]) + 7471*integer(pl[3*x])) shr 16;

    end;

  except
    on EFOpenError do Exit;
  end;
  jp.Destroy;
  BitMap.Destroy;
end; // procedure matrix.LoadGreyFromFile(fname: string);// w[y,x] := ITU-601(red,green,blue)

// -------------------------------------------------------------
procedure matrix.LoadGreyFromFile_(fname: string);// w[y,x] := (red + green + blue)/3
var BitMap : TBitMap;
    jp     : TJpegImage;
    x,y    : integer;
    ext    : string[5];
    pl     : PByteArray;
begin
  BitMap := TBitMap.Create;
  jp := TJpegImage.Create;

  ext := UpperCase( ExtractFileExt(fname));
  if (ext <> '.BMP') and (ext <> '.JPG') and (ext <> '.JPEG') then Exit;

  try
    if ext= '.BMP' then BitMap.LoadFromFile(fname)
    else begin
      jp.LoadFromFile(fname);
      BitMap.Assign(jp);
    end;
    my:=Bitmap.Height;
    mx:=Bitmap.Width;
    SetLength(w, my, mx);
    {     : Bitmap -> (pr, pg, pb)}
    for y:= 0 to my-1 do begin
      pl := BitMap.ScanLine[y];
      for x:= 0 to mx-1 do
      //w[y,x] := integer(pl[3*x  ]) + pl[3*x+1] + pl[3*x+2];
      w[y,x] := (1 + integer(pl[3*x  ]) + pl[3*x+1] + pl[3*x+2]) div 3;
    end;

  except
    on EFOpenError do Exit;
  end;
  jp.Destroy;
  BitMap.Destroy;
end; // procedure matrix.LoadGreyFromFile_(fname: string);// w[y,x] := (red + green + blue)/3

// -----------------------------------------------------------------------------
procedure LoadFromFile(r,g,b: matrix; fname: string);
var BitMap : TBitMap;
    jp     : TJpegImage;
    mx,my, x,y: integer;
    ext    : string[5];
    pl     : PByteArray;
begin
  BitMap := TBitMap.Create;
  jp := TJpegImage.Create;

  ext := UpperCase( ExtractFileExt(fname));
  if (ext <> '.BMP') and (ext <> '.JPG') and (ext <> '.JPEG') then Exit;

  try
    if ext= '.BMP' then BitMap.LoadFromFile(fname)
    else begin
      jp.LoadFromFile(fname);
      BitMap.Assign(jp);
    end;
    my:=Bitmap.Height;
    mx:=Bitmap.Width;
    r.setSize(mx, my);
    g.setSize(mx, my);
    b.setSize(mx, my);
    {     : Bitmap -> (pr, pg, pb)}
    for y:= 0 to my-1 do begin
      pl := BitMap.ScanLine[y];
      for x:= 0 to mx-1 do begin
			b.put(x,y, pl[3*x  ]);
			g.put(x,y, pl[3*x+1]);
			r.put(x,y, pl[3*x+2]);
      end;
    end;

  except
    on EFOpenError do Exit;
  end;
  jp.Destroy;
  BitMap.Destroy;
end; // procedure LoadFromFile(r,g,b: matrix; fname: string);


// -------------------------------------------------------------
procedure   matrix.SaveGreyToFile  (fname: string);
var BitMap : TBitMap;
    jp     : TJpegImage;
    x,y    : integer;
    ext    : string[5];
    pl     : PByteArray;
    b      : byte;
begin
  ext := UpperCase( ExtractFileExt(fname));
  if (ext <> '.BMP') and (ext <> '.JPG') and (ext <> '.JPEG') then Exit;
  BitMap := TBitMap.Create;
  Bitmap.PixelFormat := pf24bit;
  Bitmap.Height := my;
  Bitmap.Width  := mx;
  {     : Bitmap <- p1 }
  for y:= 0 to my-1 do begin
    pl := BitMap.ScanLine[y];
    for x:= 0 to mx-1 do begin
      b := extended2byte( w[y,x] );
      pl[3*x  ] := b;
      pl[3*x+1] := b;
      pl[3*x+2] := b;
    end;
  end;

    if ext='.BMP' then begin
        BitMap.SaveToFile(fname);
    end else begin
        jp := TJpegImage.Create;
        jp.Assign(Bitmap);
        jp.CompressionQuality:=CompressionQuality;

        try
           jp.SaveToFile(fname);
        except
           on EFOpenError do ErrProc( 'jpeg not saved');
        end;
        jp.Free;
    end;
    BitMap.Free;

end; // procedure   matrix.SaveGreyToFile  (fname: string);

// -----------------------------------------------------------------------------
procedure SaveToFile (r,g,b: matrix; fname: string);
var BitMap : TBitMap;
    jp     : TJpegImage;
    mx,my,x,y : integer;
    ext    : string[5];
    pl     : PByteArray;
begin
  r.testSameSize(g);
  b.testSameSize(g);
  mx := r.getMx;
  my := r.getMy;
  ext := UpperCase( ExtractFileExt(fname));
  if (ext <> '.BMP') and (ext <> '.JPG') and (ext <> '.JPEG') then Exit;
  BitMap := TBitMap.Create;
  Bitmap.PixelFormat := pf24bit;
  Bitmap.Height := my;
  Bitmap.Width  := mx;
  {     : Bitmap <- p1 }
  for y:= 0 to my-1 do begin
    pl := BitMap.ScanLine[y];
    for x:= 0 to mx-1 do begin
      pl[3*x  ] := b.getI(x,y);
      pl[3*x+1] := g.getI(x,y);
      pl[3*x+2] := r.getI(x,y);
    end;
  end;

    if ext='.BMP' then begin
        BitMap.SaveToFile(fname);
    end else begin
        jp := TJpegImage.Create;
        jp.Assign(Bitmap);
        jp.CompressionQuality:=CompressionQuality;

        try
           jp.SaveToFile(fname);
        except
           on EFOpenError do ErrProc( 'jpeg not saved');
        end;
        jp.Free;
    end;
    BitMap.Free;

end; // procedure SaveToFile  (r,g,b: matrix; fname: string);


// -----------------------------------------------------------------------------
procedure   matrix.SaveToTextFile  (fname: string);   // Save to text file
var ft: TextFile;
    y, x: integer;
begin
  Assign( ft, fname); Rewrite( ft );
  WriteLn( ft, mx, ' ', my );
  for y:=0 to my-1 do begin
    for x:=0 to mx-1 do
        //Write(ft, ' (',y:4,',',x:4,')=', round(w[y,x]):4, ' ');
		Write(ft, w[y,x]:20:16, ' ');
		//Write(ft, w[y,x]:16, ' ');
	WriteLn(ft);
  end;
  CloseFile(ft);
end;
// -------------------------------------------------------------
procedure   matrix.SaveToTextFile  (fname: string; l:integer);   // Save to text file
var ft: TextFile;
    y, x: integer;
begin
  Assign( ft, fname); Rewrite( ft );
  WriteLn( ft, mx, ' ', my );
  for y:=0 to my-1 do begin
    for x:=0 to mx-1 do
        Write(ft, N2Str(w[y,x], l), ' ');
    WriteLn(ft);
  end;
  CloseFile(ft);
end;
// -------------------------------------------------------------
procedure matrix.SaveToTextFileXY (fname: string; l:integer);// Save to text file
var ft: TextFile;
    y, x: integer;
begin
  Assign( ft, fname); Rewrite( ft );
  WriteLn( ft, mx, ' ', my );
  for y:=0 to my-1 do begin
    for x:=0 to mx-1 do
        Write(ft, '(',y:4,x:4,')->', N2Str(w[y,x], l), '; ');
    WriteLn(ft);
  end;
  CloseFile(ft);
end;
// -------------------------------------------------------------
procedure   matrix.SaveIToTextFile  (fname: string); // Save integers to text file
var ft: TextFile;
    y, x: integer;
begin
  Assign( ft, fname); Rewrite( ft );
  WriteLn( ft, mx, ' ', my );
  for y:=0 to my-1 do begin
    for x:=0 to mx-1 do
    	Write(ft, round(w[y,x]):4, ' ');
    	//Write(ft, '(',x:3,',',y:3,':',round(w[y,x]):5, ') ');
    WriteLn(ft);
  end;
  CloseFile(ft);
end;
// -------------------------------------------------------------
procedure   matrix.SaveIToTextFileXY(fname: string); // Save integers to text file
var ft: TextFile;
    y, x: integer;
begin
  Assign( ft, fname); Rewrite( ft );
  WriteLn( ft, mx, ' ', my );
  for y:=0 to my-1 do begin
    for x:=0 to mx-1 do
    	Write(ft, '(',x:3,',',y:3,':',round(w[y,x]):5, ') ');
    WriteLn(ft);
  end;
  CloseFile(ft);
end;
// -------------------------------------------------------------
procedure matrix.LoadFromTextFile(fname: string);   // Load from text file
var ft: TextFile;
    y, x: integer;
begin
  Assign( ft, fname); ReSet( ft );
  ReadLn( ft, mx, my );
  SetLength( w, my, mx );
  for y:=0 to my-1 do begin
    for x:=0 to mx-1 do Read(ft, w[y,x]);
    //ReadLn(ft);
  end;
  CloseFile(ft);
end;

// ------------------------------------------------------------
function    matrix.get1 (x,y: extended ):extended;    // bilinear interpolation
var y0, x0: integer;
begin
  if y<0 then y:=0; if y>(my-1) then y:=my-1;
  if x<0 then x:=0; if x>(mx-1) then x:=mx-1;
  y0 := floor(y);  x0 := floor(x);
  y  := y-y0;      x  := x-x0;
  get1 := get(x0  ,y0  )*(1-y)*(1-x) +
          get(x0+1,y0  )*(1-y)*x     +
          get(x0  ,y0+1)*y    *(1-x) +
          get(x0+1,y0+1)*y    *x     ;
end;
// ------------------------------------------------------------
function  Cubic( fm, f0, f1, f2, x: extended ): extended; // cubic interpolation (by 4 points)
begin
  Cubic :=  f0
          +(f1-1/6*f2-1/3*fm-1/2*f0)*x
          +(1/2*f1-f0+1/2*fm)*x*x
          +(1/6*f2-1/2*f1+1/2*f0-1/6*fm)*x*x*x;
end;
// ------------------------------------------------------------
function    matrix.get3 (x,y: extended ):extended;    // bicubic interpolation
var y0, x0: integer;
    fm, f0, f1, f2: extended;
begin
  if y<0 then y:=0; if y>(my-1) then y:=my-1;
  if x<0 then x:=0; if x>(mx-1) then x:=mx-1;
  y0 := floor(y);  x0 := floor(x);
  y  := y-y0;      x  := x-x0;

  fm := Cubic( get(x0-1, y0-1), get(x0, y0-1), get(x0+1, y0-1), get(x0+2, y0-1), x );
  f0 := Cubic( get(x0-1, y0  ), get(x0, y0  ), get(x0+1, y0  ), get(x0+2, y0  ), x );
  f1 := Cubic( get(x0-1, y0+1), get(x0, y0+1), get(x0+1, y0+1), get(x0+2, y0+1), x );
  f2 := Cubic( get(x0-1, y0+2), get(x0, y0+2), get(x0+1, y0+2), get(x0+2, y0+2), x );
  get3 := Cubic( fm, f0, f1, f2, y );
end;

// algebra -----------------------------------------------------------
procedure   matrix.setValue( u: extended );            // w[*.*] := x;
var x, y: integer;
begin
    for y:=0 to my-1 do
    for x:=0 to mx-1 do w[y,x]:=u;
end;

// ----------------------------------------------------------------------
procedure   matrix.addValue( u: extended );            // w[*.*] := u;
var x, y: integer;
begin
    for y:=0 to my-1 do
    for x:=0 to mx-1 do w[y,x]:=w[y,x]+u;
end;
// ----------------------------------------------------------------------
procedure   matrix.mulValue( u: extended );            // w[*.*] := u;
var x, y: integer;
begin
    for y:=0 to my-1 do
    for x:=0 to mx-1 do w[y,x]:=w[y,x]*u;
end;

// ----------------------------------------------------------------------
procedure   matrix.add( v:matrix );
var x, y: integer;
begin
    self.testSameSize(v);
    for y:=0 to my-1 do
    for x:=0 to mx-1 do w[y,x]:=w[y,x]+v.get(x,y);
end;
// ----------------------------------------------------------------------
procedure   matrix.sub( v:matrix );
var x, y: integer;
begin
    self.testSameSize(v);
    for y:=0 to my-1 do
    for x:=0 to mx-1 do w[y,x]:=w[y,x]-v.get(x,y);
end;
// ----------------------------------------------------------------------
procedure   matrix.addC( v: matrix; k: extended );     // self += v*u
var x, y: integer;
begin
    self.testSameSize(v);
    for y:=0 to my-1 do
    for x:=0 to mx-1 do w[y,x]:=w[y,x]+k*v.get(x,y);
end;
// ----------------------------------------------------------------------
procedure   matrix.mul( v: matrix );                   //  
var x, y: integer;
begin
    self.testSameSize(v);
    for y:=0 to my-1 do
    for x:=0 to mx-1 do w[y,x]:=w[y,x]*v.get(x,y);
end;

// ----------------------------------------------------------------------
function   matrix.min():extended;
var x, y: integer;
    min0: extended;
begin
    min0:= 1e999;
    for y:=0 to my-1 do
    for x:=0 to mx-1 do
    if w[y,x]<min0 then min0:=w[y,x];
    min := min0;
end;
// ----------------------------------------------------------------------
function   matrix.min(var x,y: integer):extended;
var x0, y0: integer;
    min0: extended;
begin
    min0:= 1e999;
    for y0:=0 to my-1 do
    for x0:=0 to mx-1 do
    if w[y0,x0]<min0 then begin min0:=w[y0,x0]; x:=x0; y:=y0; end;
    min := min0;
end;
// ----------------------------------------------------------------------
function   matrix.max():extended;
var x, y: integer;
    max0: extended;
begin
    max0:= -1e999;
    for y:=0 to my-1 do
    for x:=0 to mx-1 do
    if w[y,x]>max0 then max0:=w[y,x];
    max := max0;
end;
// ----------------------------------------------------------------------
function   matrix.max(var x,y: integer):extended;
var x0, y0: integer;
    max0: extended;
begin
    max0:= 1e999;
    for y0:=0 to my-1 do
    for x0:=0 to mx-1 do
    if w[y0,x0]>max0 then begin max0:=w[y0,x0]; x:=x0; y:=y0; end;
    max := max0;
end;
// ----------------------------------------------------------------------
procedure  matrix.clip( k0, k1: extended ); //      [k0,k1]
var x, y: integer;
    w0  : extended;
begin
    for y:=0 to my-1 do
    for x:=0 to mx-1 do begin
        w0 := w[y,x];
        if w0<k0 then w[y,x]:=k0;
        if w0>k1 then w[y,x]:=k1;
    end;
end;
// -----------------------------------------------------------------------------
procedure matrix.norm( k0, k1: extended );    //      [k0,k1]
var x, y: integer;
    mi, ma, w0, c0, c1  : extended;
begin
	mi := w[0,0];
    ma := mi;
    for y:=0 to my-1 do
    for x:=0 to mx-1 do begin
        w0 := w[y,x];
        if w0<mi then mi := w0;
        if w0>ma then ma := w0;
    end;
    if mi=ma then begin self.setValue((k0+k1)/2); Exit; end;
    c1 := (k1-k0)/(ma-mi);
    c0 := k0 - mi*c1;
    for y:=0 to my-1 do
    for x:=0 to mx-1 do begin
        w0 := w[y,x];
        w[y,x]:=c0 + c1*w0;
    end;
end;

// -----------------------------------------------------------------------------
procedure  matrix.roundI();
var x, y: integer;
begin
    for y:=0 to my-1 do
    for x:=0 to mx-1 do
    w[y,x] := round(w[y,x]);
end;
// -----------------------------------------------------------------------------
//     a[k-n..k+n]
function middleValue( a: array of extended; size, n, k: integer): extended;
var i: integer;
    s: extended;
begin
    s:=0;
    if k<n then begin   //   
        for i:=0 to k+n do s := s+a[i];
        s := s/(k+n+1);
        middleValue := s;
        exit;
    end;

    if k+n>size-1 then begin    //  
        for i:=k-n to size-1 do s := s+a[i];
        s := s/(size-k+n);
        middleValue := s;
        exit;
    end;

    //  
    for i:=k-n to k+n do s := s+a[i];
    s := s/(2*n+1);
    middleValue := s;
end;
// ----------------------------------------------------------------------
//     a[k-n..k+n]
function middleValue2( a: array of extended; size, n, k: integer): extended;
var i, cnt: integer;
    s: extended;
begin
    if (k<0) or (k>=size) then ErrProc( 'middleValue2: k='+IntToStr(k));
    if (n<1) or (n>=size/2)then ErrProc('middleValue2: n='+IntToStr(n));

    s := n*a[k]; cnt := n;
    for i:=1 to n-1 do begin
        if k-i>=0   then begin s := s+ (n-i)*a[k-i]; cnt := cnt+n-i; end;
        if k+i<size then begin s := s+ (n-i)*a[k+i]; cnt := cnt+n-i; end;
    end;
    s := s/cnt;
    middleValue2 := s;

end;
// ----------------------------------------------------------------------
procedure   matrix.smoothSlow(n: integer);   //  :    -     (2k+1)*(2k+1)
var buf: array of extended;
    x, y, size: integer;
begin
	if n=0 then Exit;
	//WriteLn('matrix.smooth(',n,')');
    size := mx; if my>mx then size:=my;
    setLength( buf, size);
    //    y:
    for y:=0 to my-1 do begin
        for x:=0 to mx-1 do buf[x] := self.get(x,y);    //    
        for x:=0 to mx-1 do self.put(x,y, middleValue(buf,mx,n,x) );
        if (y mod 100)=0 then Write(' y=', y:4, '/', my:4, ' '#13 );
    end;
    //   x:
    for x:=0 to mx-1 do begin
        for y:=0 to my-1 do buf[y] := self.get(x,y);    //    
        for y:=0 to my-1 do self.put(x,y, middleValue(buf,my,n,y) );
        if (x mod 100)=0 then Write(' x=', x:4, '/', mx:4, ' '#13 );
    end;
    buf:=NIL;
end;
// ----------------------------------------------------------------------
procedure   matrix.smooth(n: integer);         //    -     (2k+1)*(2k+1)
var buf: array of extended;
    x, y, size: integer;
    su, co, t: extended;
begin
	if n=0 then Exit;
	//WriteLn('matrix.smooth(',n,')');
    size := mx; if my>mx then size:=my; size := size+2*n+2;
    setLength( buf, size);  // for x:=0 to size-1 do buf[x] := 0;
    co := 1./((2*n+1)*(2*n+1));
    //    y:
    for y:=0 to my-1 do begin
    	//    
        t:=self.w[y,   0]; for x:=0 to n do buf[       x] := t;
        su:=t*(n+1); //    su
        t:=self.w[y,mx-1]; for x:=0 to n do buf[mx-1+n+x] := t;
        for x:=1 to mx-2 do buf[x+n] := self.w[y,x];   //    

        //    su:
        for x:=1 to n do su := su+buf[x+n];

        for x:=0 to mx-1 do begin
			self.w[y,x]:=su;
            su := su - buf[x] + buf[x+2*n+1];
        end;
        if (y mod 100)=0 then Write(' y=', y:4, '/', my:4, ' '#13 );
    end;

    //   x:
    for x:=0 to mx-1 do begin
    	//    
        t:=self.w[0,   x]; for y:=0 to n do buf[       y] := t;
        su:=t*(n+1); //    su
        t:=self.w[my-1,x]; for y:=0 to n do buf[my-1+n+y] := t;
        for y:=1 to my-2 do buf[y+n] := self.w[y,x];   //    

        //    su:
        for y:=1 to n do su := su+buf[y+n];

        for y:=0 to my-1 do begin
			self.w[y,x]:=su*co;
            su := su - buf[y] + buf[y+2*n+1];
        end;

        if (x mod 100)=0 then Write(' x=', x:4, '/', mx:4, ' '#13 );
    end;
    buf:=NIL;
end;
// ----------------------------------------------------------------------
procedure   matrix.smooth2(n: integer);        //    -    ....
var buf: array of extended;
    x, y, size: integer;
begin
	if n=0 then Exit;
    size := mx; if my>mx then size:=my;
    setLength( buf, size);
    //    y:
    for y:=0 to my-1 do begin
        for x:=0 to mx-1 do buf[x] := self.get(x,y);    //    
        for x:=0 to mx-1 do self.put(x,y, middleValue2(buf,mx,n,x) );
        if (y mod 100)=0 then Write(' y=', y:4, '/', my:4, ' '#13 );
    end;
    //   x:
    for x:=0 to mx-1 do begin
        for y:=0 to my-1 do buf[y] := self.get(x,y);    //    
        for y:=0 to my-1 do self.put(x,y, middleValue2(buf,my,n,y) );
        if (x mod 100)=0 then Write(' x=', x:4, '/', mx:4, ' '#13 );
    end;
    buf:=NIL;
end;
// ----------------------------------------------------------------------
function    matrix.dist2( a: matrix ):extended; //  
var s: extended;
    x,y: integer;
begin
    testSameSize(a);
    s:=0;
    for y:=0 to my-1 do
    for x:=0 to mx-1 do begin
        s:=s+sqr( w[y,x]-a.w[y,x] );
    end;
    dist2:= sqrt( s/(mx*my) );
end;

// ----------------------------------------------------------------------
// matrix as mask
//    [x,y],   a.w[y,x]=<v,  self.w[y,x]:=m
function    matrix.setLess( a: matrix; v, m: extended ):integer;
var x,y, cnt: integer;
begin
    self.setSize(a.mx, a.my);
    cnt := 0;
    for y:=0 to my-1 do
    for x:=0 to mx-1 do
    if a.w[y,x]<=v then begin w[y,x]:=m; inc(cnt); end;
    setLess:=cnt;
end;
// ----------------------------------------------------------------------
//    [x,y],   a.w[y,x]>=v,  self.w[y,x]:=m
function    matrix.setGreater( a: matrix; v, m: extended ):integer;
var x,y, cnt: integer;
begin
    self.setSize(a.mx, a.my);
    cnt := 0;
    for y:=0 to my-1 do
    for x:=0 to mx-1 do
    if a.w[y,x]>=v then begin w[y,x]:=m; inc(cnt); end;
    setGreater:=cnt;
end;
// ----------------------------------------------------------------------
//    [x,y],   v1<=a.w[y,x]<=v2,  self.w[y,x]:=m
function    matrix.setInterval( a: matrix; v1, v2, m: extended ):integer;
var x,y, cnt: integer;
begin
    self.setSize(a.mx, a.my);
    cnt := 0;
    for y:=0 to my-1 do
    for x:=0 to mx-1 do
    if (v1<=a.w[y,x]) and (a.w[y,x]<=v2) then begin w[y,x]:=m; inc(cnt); end;
    setInterval:=cnt;
end;
// ----------------------------------------------------------------------
function    matrix.countValue( m: extended ):integer; //     
var x,y, cnt: integer;
begin
    cnt := 0;
    for y:=0 to my-1 do
    for x:=0 to mx-1 do
    if w[y,x]=m then inc(cnt);
    countValue:=cnt;
end;
// -----------------------------------------------------------------------------
function    matrix.changeValue( m1, m2: extended ):integer; //  m1     m2.   
var x,y, cnt: integer;
begin
    cnt := 0;
    for y:=0 to my-1 do
    for x:=0 to mx-1 do
    if w[y,x]=m1 then begin inc(cnt); w[y,x]:=m2; end;
    changeValue:=cnt;
end;
// -----------------------------------------------------------------------------
procedure  matrix.diffX ( a: matrix );			// this := da/dx   
var x,y: integer;
begin
	self.setSize(a.mx,a.my);
    for y:=0 to my-1 do begin
    	put(0,y, (-1.5*a.get(0,y)+2*a.get(1,y)-0.5*a.get(2,y)) );			// x=0
	    for x:=1 to mx-2 do put(x,y, (a.get(x+1,y)-a.get(x-1,y))/2);
    	put(mx-1,y, (-1.5*a.get(mx-1,y)+2*a.get(mx-2,y)-0.5*a.get(mx-3,y)));//x=mx-1
    end;
end;
// -----------------------------------------------------------------------------
procedure  matrix.diffY ( a: matrix );			// this := da/dy   
var x,y: integer;
begin
	self.setSize(a.mx,a.my);
    for x:=0 to mx-1 do begin
    	put(x,0, (-1.5*a.get(x,0)+2*a.get(x,1)-0.5*a.get(x,2)) );			 // y=0
	    for y:=1 to my-2 do put(x,y, (a.get(x,y+1)-a.get(x,y-1))/2);
    	put(x,my-1, (-1.5*a.get(x, my-1)+2*a.get(x,my-2)-0.5*a.get(x,my-3)));//y=my-1
    end;
end;
// -----------------------------------------------------------------------------
procedure  matrix.diffXX( a: matrix );          // this := da/dxdx   
var x,y: integer;
begin
	self.setSize(a.mx,a.my);
    for y:=0 to my-1 do
	for x:=0 to mx-1 do put(x,y, a.get(x+1,y)-2*a.get(x,y)+a.get(x-1,y));
end;

// -----------------------------------------------------------------------------
procedure  matrix.diffXY( a: matrix );          // this := da/dxdy   
var x,y: integer;
begin
	self.setSize(a.mx,a.my);
    for y:=0 to my-1 do
    for x:=0 to mx-1 do
    put(x,y, (a.get(x+1,y+1)-a.get(x+1,y-1)-a.get(x-1,y+1)+a.get(x-1,y-1))/4);
end;
// -----------------------------------------------------------------------------
procedure  matrix.diffYY( a: matrix );          // this := da/dydy   
var x,y: integer;
begin
	self.setSize(a.mx,a.my);
    for y:=0 to my-1 do
    for x:=0 to mx-1 do put(x,y, a.get(x,y+1)-2*a.get(x,y)+a.get(x,y-1));
end;

// =============================================================================
{
         
(  "A"  "B").

    ,    a[0<=x<mx, 0<=y<my]
    f   ,  
 [0<=x<=mx, 0<=y<=my].

  "A"  ,  a[x,y]    f(x,y),
  "B" - a[x,y]    f(x,y)   [x..x+1,y..y+1],
     .

      / 
      .  a1 -  ,
a2 -    .

+----------------------+--------------------+----------------------------------------------------+
|	           |  "A"			|  "B"                                            |
|    	   | a2[x,y]=a1[2x,2y]	| a2[x,y] =                                          |
|					   | 					| (a1[2x,2y]+a1[2x+1,2y]+a1[2x,2y+1]+a1[2x+1,2y+1])/4|
+----------------------+--------------------+----------------------------------------------------+
|   2   | a1[x,y]=a2[x/2,y/2]| a1[x,y] = a2[x/2-1/4, y/2-1/4]                     |
+----------------------+--------------------+----------------------------------------------------+

   -   .   
   "A",    , ,
 -    "B".

}
// =============================================================================
// resize "A"
// -----------------------------------------------------------------------------
//  a  times    .
//        align
procedure matrix.downSampl_A(a: matrix; times: integer; align: integer=1); // this:=downsampe(a) by times
var
	x1, y1: integer;
begin
	self.setSize(align*(a.mx div (align*times)), align*(a.my div (align*times)));
    for y1:=0 to my-1 do
    for x1:=0 to mx-1 do self.put(x1,y1, a.get(times*x1, times*y1));
end;

// -----------------------------------------------------------------------------
procedure matrix.resizeBilin_A(a: matrix; mx_, my_: integer); // this:=resize(a) to (mx_, my_)
var
	x1, y1: integer;
    cox, coy: extended;
begin
	self.setSize(mx_,my_);
    cox := 1.0*a.getMx/mx_;
    coy := 1.0*a.getMy/my_;

    for y1:=0 to my-1 do
    for x1:=0 to mx-1 do self.put(x1,y1, a.get1(cox*x1, coy*y1));

end;

// -----------------------------------------------------------------------------
procedure matrix.resizeBicub_A(a: matrix; mx_, my_: integer); // this:=resize(a) to (mx_, my_)
var
	x1, y1: integer;
    cox, coy: extended;
begin
	self.setSize(mx_,my_);
    cox := 1.0*a.getMx/mx_;
    coy := 1.0*a.getMy/my_;
    for y1:=0 to my-1 do
    for x1:=0 to mx-1 do self.put(x1,y1, a.get3(cox*x1, coy*y1));

end;


// =============================================================================
// resize "B"
// -----------------------------------------------------------------------------
//  a  times    .
//        align
procedure matrix.downSampl(a: matrix; times: integer; align: integer=1); // this:=downsampe(a) by times
var
	x,y, x1, y1, k2: integer;
    s: extended;
begin
	self.setSize(align*(a.mx div (align*times)), align*(a.my div (align*times)));
	k2 := times*times;
    for y1:=0 to my-1 do
    for x1:=0 to mx-1 do begin
		s:=0;
	    for y:=times*y1 to times*y1+times-1 do
	    for x:=times*x1 to times*x1+times-1 do s:=s+a.get(x,y);
        self.put(x1,y1, s/k2);
    end;
end;

// -----------------------------------------------------------------------------
//  a  times       ,     .
//        align
procedure matrix.downSamplI(a: matrix; times: integer; align: integer=1); // this:=downsampe(a) by times
var
	x,y, x1, y1, s, k2, bias: integer;
begin
	self.setSize(align*(a.mx div (align*times)), align*(a.my div (align*times)));
	k2 := times*times;
	bias := (k2-1) div 2;
    for y1:=0 to my-1 do
    for x1:=0 to mx-1 do begin
		s:=bias;
	    for y:=times*y1 to times*y1+times-1 do
	    for x:=times*x1 to times*x1+times-1 do s:=s+a.getI(x,y);
        self.put(x1,y1, s div k2  );
        bias := k2-1-bias;
    end;
end;

// -----------------------------------------------------------------------------
procedure matrix.resizeBilin(a: matrix; mx_, my_: integer); // this:=resize(a) to (mx_, my_)
var
	x1, y1: integer;
    cx1, cx0, cy1, cy0: extended;
begin
	self.setSize(mx_,my_);
    cx1 := 1.0*a.getMx/mx_;
    cx0 := cx1*0.5 - 0.5;

    cy1 := 1.0*a.getMy/my_;
    cy0 := cy1*0.5 - 0.5;

    for y1:=0 to my-1 do
    for x1:=0 to mx-1 do self.put(x1,y1, a.get1(cx1*x1+cx0, cy1*y1+cy0));

end; // procedure matrix.resizeBilin(a: matrix; mx_, my_: integer); // this:=resize(a) to (mx_, my_)


// -----------------------------------------------------------------------------
procedure matrix.resizeBicub(a: matrix; mx_, my_: integer); // this:=resize(a) to (mx_, my_)
var
	x1, y1: integer;
    cx1, cx0, cy1, cy0: extended;
begin
	self.setSize(mx_,my_);
    cx1 := 1.0*a.getMx/mx_;
    cx0 := cx1*0.5 - 0.5;

    cy1 := 1.0*a.getMy/my_;
    cy0 := cy1*0.5 - 0.5;

    for y1:=0 to my-1 do
    for x1:=0 to mx-1 do self.put(x1,y1, a.get3(cx1*x1+cx0, cy1*y1+cy0));

end; // procedure matrix.resizeBicub(a: matrix; mx_, my_: integer); // this:=resize(a) to (mx_, my_)

// -----------------------------------------------------------------------------
procedure matrix.expand2_Base(a: matrix); //  
var
	x, y: integer;
begin
	self.setSize(2*a.getMx, 2*a.getMy);

    for y:=0 to my-1 do
    for x:=0 to mx-1 do self.put(x,y, a.get(x div 2, y div 2));

end; // procedure matrix.expand2_Base(a: matrix); //  

// -----------------------------------------------------------------------------
procedure matrix.expand2_Bilin(a: matrix); //  
var
	x, y: integer;
begin
	self.setSize(2*a.getMx, 2*a.getMy);

    for y:=0 to my-1 do
    for x:=0 to mx-1 do self.put(x,y, a.get1(0.5*x - 0.25 , 0.5*y - 0.25));

end; // procedure matrix.expand2_Bilin(a: matrix); //  

// -----------------------------------------------------------------------------
procedure matrix.expand2_Bicub(a: matrix); //  
var
	x, y: integer;
begin
	self.setSize(2*a.getMx, 2*a.getMy);

    for y:=0 to my-1 do
    for x:=0 to mx-1 do self.put(x,y, a.get3(0.5*x - 0.25 , 0.5*y - 0.25));

end; // procedure matrix.expand2_Bicub(a: matrix); //  

// -----------------------------------------------------------------------------
procedure matrix.expand2_4Point(m: matrix; al,be,ga:extended); // 4-    (al,be,ga,1-al-be-ga)
var
	x, y, x2,x2a, y2,y2a: integer;
	de, a,b,c,d: extended;
begin
	self.setSize(2*m.getMx, 2*m.getMy);
    de := 1-al-be-ga;

    for y:=0 to my-1 do
    for x:=0 to mx-1 do begin
    	x2 := x div 2;
        if odd(x) then x2a := x2+1 else x2a:=x2-1;
    	y2 := y div 2;
        if odd(y) then y2a := y2+1 else y2a:=y2-1;
    	a := m.get(x2 ,y2 );
    	b := m.get(x2a,y2 );
    	c := m.get(x2 ,y2a);
    	d := m.get(x2a,y2a);
		self.put(x,y, al*a + be*b + ga*c + de*d );
    end;
end; // procedure matrix.expand2_4Point(m: matrix; al,be,ga:extended); // 4-    (al,be,ga,1-al-be-ga)

// -----------------------------------------------------------------------------
begin
    CompressionQuality := 95;
end.


